// ================================================================
//
// CBP_Pioneer3_Sick_2.c
// 05-11-20 RELEASE
//
// Braitenberg light with obstacle avoidance using Lidar
//
// =================================================================

#include <webots/display.h>
#include <webots/lidar.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/light_sensor.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

# include "CBP_Plotter_2.h"

#define TIME_STEP 16
#define MAX_SPEED 6.28
#define PI 3.14159265358979323846
#define RANGE_THRESHOLD 2.0

#define DEBUG false

WbDeviceTag lidar;
WbDeviceTag left_wheel;
WbDeviceTag right_wheel;
WbDeviceTag display;
WbDeviceTag comp;
WbDeviceTag leftLight;
WbDeviceTag rightLight;
WbImageRef background;
  
int lidar_width;
static int px[180];

// attractive and repulsive factors
double k_att = 50.0;
double k_rep = 0.5;

// maximum distance of influence for the obstacles
double d_0 = 1.0;

// Display SICK laser data
static void update_display(WbDeviceTag display, WbImageRef background,float *scanvalue, int laser_width) {
  int half_width2 = laser_width / 2;
  int half_height2 = laser_width / 2;
  // (re)paint background
  wb_display_image_paste(display, background, 0, 0, false);
  // draw polygon scannings
  int py[laser_width];
  int i;
  for (i = 0; i < laser_width; i++) {
    if (scanvalue[i] > 8)  // range up to 8metres
      scanvalue[i] = 8;
    px[i] = half_width2 - (cos(i * PI / 180)) * ((90 * scanvalue[i]) / 8);
    py[i] = half_height2 - (sin(i * PI / 180)) * ((90 * scanvalue[i]) / 8);
  }
  // update Scan Display
  wb_display_draw_polygon(display, px, py, 180);
  
}

static double compute_force_obstacle(double dObj) {

  double force = 0;
  
  // Calculate the repulsive force here
  
  return force;

}

void init() {

  // get devices
  lidar = wb_robot_get_device("Sick LMS 291");
  left_wheel = wb_robot_get_device("left wheel");
  right_wheel = wb_robot_get_device("right wheel");
  display = wb_robot_get_device("display");
  comp = wb_robot_get_device("compass");
//  wb_compass_enable(comp, TIME_STEP);
  
  // init lidar
  wb_lidar_enable(lidar, TIME_STEP);
  lidar_width = wb_lidar_get_horizontal_resolution(lidar);
  int max_range = wb_lidar_get_max_range(lidar);
  printf("Lidar width = %d, lidar maxRanges = %d\n",lidar_width,max_range);
  wb_lidar_enable_point_cloud(lidar);
  printf("Sick width = %d\n",lidar_width);
  

  // Light Sensors
  leftLight = wb_robot_get_device("lsLeft");
  wb_light_sensor_enable(leftLight, TIME_STEP); 
  rightLight = wb_robot_get_device("lsRight");
  wb_light_sensor_enable(rightLight, TIME_STEP); 
    
  // init the display
  wb_display_set_color(display, 0x0000ff);
  background =  wb_display_image_load(display, "grid_display.png");

  // init motors
  wb_motor_set_position(left_wheel, INFINITY);
  wb_motor_set_position(right_wheel, INFINITY);

}

int main(int argc, char **argv) {

  double thetaObj;
  double omegaRobot;
  double omegaRobotLight; 
  double omegaRobotForce; 
  double forceRep;
  float range;
  int count;
  double omegaL,omegaR;
  
  wb_robot_init();

  init();
  
  // Grab Controller arguments
  k_att = atof(argv[1]);
  k_rep = atof(argv[2]);
  d_0   = atof(argv[3]);
  
  for(int i = 0;i<argc;i++)
    printf("argv[%d] = %s\n",i,argv[i]);
    
while (wb_robot_step(TIME_STEP) != -1) {
  
  // get lidar values
  const float *lidar_values = wb_lidar_get_range_image(lidar);

  update_display(display, background, lidar_values, lidar_width);

  // get light values and set thetaRobot
  double leftIntensity  = wb_light_sensor_get_value(leftLight);
  double rightIntensity = wb_light_sensor_get_value(rightLight); 
  
  // Calculate omegaRobotLight to turn the robot to face the light
  //omegaRobotLight = 
  
  
  if(DEBUG) printf("Left light = %f, Right light = %f\n",leftIntensity,rightIntensity);
  
  // Compute forces here -----------------------------------------------
  printf(">>>>>\n");
  count = 0;
  forceRep = 0;
  thetaObj = 0;
   for(int i = 0;i<lidar_width;i++) {
     range = lidar_values[i];





 
   } 

 
//   if(DEBUG) printf("AvTheta = %f\n",avTheta);
   
   avTheta -= 1.0*M_PI/180.0; 	// correction for LIDAR angle offset
   // Now use the force to set the anglular velocity of the robot
   if(count > 0) {






    }
  
    double velyRobot = 0.5*MAX_SPEED;
    
    omegaRobot = omegaRobotForce + omegaRobotLight;
    
    printf("omegaRobot = %f, omegaRLight = %f\n",omegaRobot,omegaRobotLight);
    
    omegaR = velyRobot + omegaRobot;
    omegaL = velyRobot - omegaRobot;
    
    // Get this condition from the LUT values
    if(leftIntensity > 195 || rightIntensity > 195){
      omegaL = 0;
      omegaR = 0;
    }

    wb_motor_set_velocity(left_wheel, omegaL);
    wb_motor_set_velocity(right_wheel, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}
