// =================================================================
//
// CBP_ePuck_Camera_3.c 
//
// Converts RGB to HSI and segments red pixels
// Rotates robot to centralize red object
// CBP 16-10-20
// RELEASE
//
// ==================================================================

#include <stdio.h>
#include <webots/camera.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/display.h>
#include <webots/distance_sensor.h>

#include <stdlib.h>

#include "CBP_Helper_4.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979
#define THRESHOLD_DIST 0.01

#define STATE_SEARCH 1
#define STATE_FOUND 2
#define STATE_MOVETO_TARGET 3
#define STATE_STOP 4

static WbDeviceTag display1;   
WbDeviceTag toflight;
WbDeviceTag camera, left_motor, right_motor;
int width, height;
 
static unsigned char *image1;


int imageSegment(const unsigned char *image,unsigned char *imageThresh, int width, int height) {
  
  int count = 0;
  double minsat = 1.5;
  double maxsat = 0;
  for (int i = 0; i < width ; i++) {
    for (int j = 0; j < height; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      double green = wb_camera_image_get_green(image, width, i, j);
      double blue = wb_camera_image_get_blue(image, width, i, j);  
      
      // Find and remove white light pollution      
      double min = 255;
      if(red < min)
        min = red;
      if(green < min)
        min = green;
      if(blue < min)
        min = blue;
       double intens = red + green + blue + 0.000001;
       
       double sat = (intens - min)/intens;
       if(sat < minsat)
         minsat = sat;
       if(sat > maxsat)
         maxsat = sat;         
      red -= min;
      blue -= min;
      green -= min;
                     
      double sum = red + green + blue + 0.000001;
      red = red/sum;
      green = green/sum;
      blue = blue/sum;
      
      double numi=0.5*((red - green)+(red - blue));
      double denom=sqrt((red-green)*(red - green)+ (red-blue)*(green - blue));
      double hue = acos(numi/(denom + 0.0000001))*180/M_PI;
      if(blue > green) hue = 360-hue;   
      //printf("Sat = %f\n",sat);
      // 4 bytes per pixel in the image. Offset 1 is red
      *(imageThresh + i*4 + j*width*4) = 0;
      *(imageThresh + i*4+1 + j*width*4) = 0;
      *(imageThresh + i*4+2 + j*width*4) = 0;
      *(imageThresh + i*4+3 + j*width*4) = 0;      
      if(( hue > 0 && hue < 40)  || (hue > 330 && hue < 360 && sat > 0.7)){
      //if(( hue > 115 && hue < 125) && sat > 0.7){      
         *(imageThresh + i*4 + 1 + j*width*4) = 250;
        count++;
      }
    }
        
  }
  printf("Min sat = %f, max sat = %f\n",minsat,maxsat);
  return count;
  
}


int getAvRedLocation(unsigned char *image, int width, int height){

   int indexSum = 0;
   int indexAv = 0;
   int count = 0;
   for (int i = 0; i < width ; i++) { 
     if(*(image + i*4 + 1 + (height/2)*width*4) == 250) {
       indexSum += i; 
       count++;
     }
   }
   printf("IndexSum = %d\n",indexSum);
   indexAv = indexSum/count;
   return indexAv;   
} 

void init() {

  camera = wb_robot_get_device("camera");
  wb_camera_enable(camera, TIME_STEP);
  width = wb_camera_get_width(camera);
  height = wb_camera_get_height(camera);
  printf("Camera width: %d height: %d\n",width,height);

  /* get a handler to the motors and set target position to infinity (speed control). */
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  // get the display device
  display1 = wb_robot_get_device("display");
  width = wb_display_get_width(display1); 
  height = wb_display_get_height(display1);
 
  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP);
 
}


int main() {

  double omegaL, omegaR;

  int state = STATE_SEARCH;

  wb_robot_init();

  init();
 
  wb_robot_step(TIME_STEP);
  
  image1 = (unsigned char *) malloc(sizeof(unsigned char)*width*height*8);
  
  /* Main loop */
  while (wb_robot_step(TIME_STEP) != -1) {
	  
    /* Get the new camera values */
    const unsigned char *image = wb_camera_get_image(camera);
    int nr_red = imageSegment(image,image1,width,height);
    WbImageRef iref = wb_display_image_new(display1, width, height, image1, WB_IMAGE_BGRA);
    wb_display_image_paste(display1, iref, 0, 0, false);
   

    switch(state) {
      case STATE_SEARCH:
        omegaL = -0.25*MAX_SPEED;
        omegaR =  0.25*MAX_SPEED;
        // Transit to STATE_FOUND
        break;
      case STATE_FOUND:
        printf(">>> STATE_FOUND object\n");
        omegaL = -0.1*MAX_SPEED;
        omegaR =  0.1*MAX_SPEED; 
        printf("omL = %f\n",omegaL);
        // Get centre of the object
        
        // Test if object is to left or right of centre.

        // Set omegaL and omegaR accordingly
        
        // Now transit to MOVETO_TARGET when the object is centered
        break; 
      case STATE_MOVETO_TARGET:
        omegaL = 0.75*MAX_SPEED;
        omegaR = 0.75*MAX_SPEED;
        // Check time of flight and stop when close
        double sensorVal = wb_distance_sensor_get_value(toflight);
        double dist = lookUpDistanceTOF(cTOF,sensorVal);
        printf("sensorval = %f, dist = %f\n",sensorVal,dist);
        if(dist < THRESHOLD_DIST)
          state = STATE_STOP;
        break;
      case STATE_STOP:
        omegaL = 0;
        omegaR = 0;
        break;        
    }

    wb_motor_set_velocity(left_motor, omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}
