// =================================================================
//
// CBP_ePuck_Camera_3a.c 
//
// Red value from RGB displayed in a separate window
// Show that R-value cannot be used to indentify red objects
// CBP 17-10-20
// RELEASE
//
// ==================================================================

#include <stdio.h>
#include <webots/camera.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/display.h>
#include <webots/distance_sensor.h>

#include <stdlib.h>

#include "CBP_Helper_4.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

static WbDeviceTag display1;   
WbDeviceTag toflight;
WbDeviceTag camera, left_motor, right_motor;
int width, height;
 
static unsigned char *image1;

int imageSegment_bad(const unsigned char *image,unsigned char *imageThresh, int width, int height) {
  for (int i = 0; i < width ; i++) {
    for (int j = 0; j < height; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      *(imageThresh + i*4 + 1 + j*width*4) = (unsigned char)red;   
    }
  }
  return -1;
}


void init() {

  camera = wb_robot_get_device("camera");
  wb_camera_enable(camera, TIME_STEP);
  width = wb_camera_get_width(camera);
  height = wb_camera_get_height(camera);
  printf("Camera width: %d height: %d\n",width,height);

  /* get a handler to the motors and set target position to infinity (speed control). */
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  // get the display device
  display1 = wb_robot_get_device("display");
  width = wb_display_get_width(display1); 
  height = wb_display_get_height(display1);
 
  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP);
 
}

int main() {

  double omegaL, omegaR;

  wb_robot_init();

  init();
 
  wb_robot_step(TIME_STEP);
  
  image1 = (unsigned char *) malloc(sizeof(unsigned char)*width*height*8);
  
  /* Main loop */
  while (wb_robot_step(TIME_STEP) != -1) {
	  
    /* Get the new camera values */
    const unsigned char *image = wb_camera_get_image(camera);
    int nr_red = imageSegment_bad(image,image1,width,height);
    WbImageRef iref = wb_display_image_new(display1, width, height, image1, WB_IMAGE_BGRA);
    wb_display_image_paste(display1, iref, 0, 0, false);
    printf("Nr red = %d\n",nr_red);

    omegaL = -0.25*MAX_SPEED;
    omegaR =  0.25*MAX_SPEED;
   
    wb_motor_set_velocity(left_motor, omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}
