// =================================================================
//
// CBP_ePuck_Camera_4a.c 
//
// CBP 24-10-20
// Uses image processing to navigate between multiple obstacles
//
// ==================================================================


#include <stdio.h>
#include <webots/camera.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/display.h>
#include <webots/distance_sensor.h>

#include <stdlib.h>

#include "CBP_Helper_4.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

#define STATE_MOVE_FORWARD 1
#define STATE_OBSTACLE_AVOID 2

static WbDeviceTag display1;   
WbDeviceTag toflight;
  WbDeviceTag camera, left_motor, right_motor;

  int height,width;
    
static unsigned char *image1;

int imageSegment(const unsigned char *image,unsigned char *imageThresh, int width, int height) {
  
  int count = 0;
  double minsat = 1.5;
  double maxsat = 0;
  for (int i = 0; i < width ; i++) {
    for (int j = 0; j < height; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      double green = wb_camera_image_get_green(image, width, i, j);
      double blue = wb_camera_image_get_blue(image, width, i, j);  
      
      // Find and remove white light pollution      
      double min = 255;
      if(red < min)
        min = red;
      if(green < min)
        min = green;
      if(blue < min)
        min = blue;
       double intens = red + green + blue + 0.000001;
       
       double sat = (intens - min)/intens;
       if(sat < minsat)
         minsat = sat;
       if(sat > maxsat)
         maxsat = sat;         
      red -= min;
      blue -= min;
      green -= min;
                     
      double sum = red + green + blue + 0.000001;
      red = red/sum;
      green = green/sum;
      blue = blue/sum;
      
      double numi=0.5*((red - green)+(red - blue));
      double denom=sqrt((red-green)*(red - green)+ (red-blue)*(green - blue));
      double hue = acos(numi/(denom + 0.0000001))*180/M_PI;
      if(blue > green) hue = 360-hue;   
      //printf("Sat = %f\n",sat);
      // 4 bytes per pixel in the image. Offset 1 is red
      *(imageThresh + i*4 + j*width*4) = 0;
      *(imageThresh + i*4+1 + j*width*4) = 0;
      *(imageThresh + i*4+2 + j*width*4) = 0;
      *(imageThresh + i*4+3 + j*width*4) = 0;      
      if(( hue > 0 && hue < 40)  || (hue > 330 && hue < 360 && sat > 0.7)){
         *(imageThresh + i*4 + 1 + j*width*4) = 250;
        count++;
      }
    }
        
  
  }
  printf("Min sat = %f, max sat = %f\n",minsat,maxsat);
  return count;
  
}

// Scans the red image looks for the tallest object. Assuming all objects are the
// same size, this will be the closest object. Function returns the height of
// the tallest object and also the centre index of the object. Uses the middle
// row through the red image.
int getBiggestRedBlob(unsigned char *image, int width, int height, double *ht){

  int blobs[10];
  int widths[10];
  int heights[10];
  int blobNr = 0;
  int count = 0;
  
  for(int i=0;i<10;i++){
    blobs[i] = 0;
    widths[i] = 0;
    heights[i] = 0;
  }
  blobNr = 0;
  count = 0;
  
  int i=1;
  // Scan the image slice (centre row) from left to right
  // Note the j-value is replaced by (height/2)
  while(i < width) {

    unsigned char prev = *(image + (i-1)*4 + 1 + (height/2)*width*4);
    unsigned char curr = *(image + (i)*4 + 1 + (height/2)*width*4);

    // Skip over the non-red pixels - gaps between objects
    while(curr == 0 && i < width) {
      curr = *(image + (i)*4 + 1 +(height/2)*width*4);
      i++;
    } 
  
    // Scan over a stretch of red pixels, store the number of pixels in the
    // stretch and the sum of the pixel indices so we can calculate the centre of the object
    while(curr == 250 && i < width) {
      blobs[blobNr] += i;  // Add current index to blobs array
      count++;
      widths[blobNr] = count;
      curr = *(image + (i)*4 + 1 + (height/2)*width*4);
      prev = *(image + (i-1)*4 + 1 + (height/2)*width*4);  
      if(prev == 250 && curr == 0) {
        count = 0;
        blobNr++;  
      }           
      i++;
      // Here we deal with any red stretch at the right of the image which does not
      // have a black stretch beyound it
      if(curr == 250 && i >= width) {
         count = 0;
        blobNr++;  
      }       
    }
    
  }
  // So here we have finishes scanning the camera image slice and we
  // found blobNr blobs and have stored wither widths
     
   // Scan over the blobs and for each blob get the height of the blob
   for(int i = 0;i<blobNr;i++){  
     // First get centre of blobs - divide sum of indices by blob with
     int centre = (int)((double)blobs[i]/(double)widths[i]);
     // No scan the blob colum and get its height in pixels
     for(int j=0; j<height;j++) {
        unsigned char pixel = *(image + centre*4 + 1 + j*width*4);
        if(pixel == 250)
          heights[i] +=1;     
     }
   }

  // find the tallest object
  double maxHeight = 0;
  int blobFound = -1;
  for(int i = 0;i<blobNr;i++){
    if(heights[i] > maxHeight) {
      maxHeight = heights[i];
      blobFound = i;
    }
  }

  // dump the objects
  for(int i = 0;i<blobNr;i++) {
    printf("Blob %d sum %d width %d height %d avCol %f\n",
     i,blobs[i],widths[i],heights[i],(double)blobs[i]/(double)widths[i]);
  }  
            
   // Return the height (ht) and the centre index for the tallest object         
   if(blobFound != -1) {
     printf("Largest blob has width %d and locn %f\n",widths[blobFound],(double)blobs[blobFound]/(double)widths[blobFound]);
     *ht = heights[blobFound];
     printf("ht %f\n",*ht);
     return (int)((double)blobs[blobFound]/(double)widths[blobFound]); // ventre index
   } else {
     *ht = 0;
     return -1;
   }
}


void init() {

  /* Get the camera device, enable it, and store its width and height */
  camera = wb_robot_get_device("camera");
  wb_camera_enable(camera, TIME_STEP);
  width = wb_camera_get_width(camera);
  height = wb_camera_get_height(camera);
  printf("Camera width: %d height: %d\n",width,height);

  /* get a handler to the motors and set target position to infinity (speed control). */
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  display1 = wb_robot_get_device("display");
  width = wb_display_get_width(display1); 
  height = wb_display_get_height(display1);

  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP);
 
}

int main() {

  double omegaL, omegaR;
  double blobHeight;
  double alpha;
  double omegaRobot = 0.0;
  double gain;
  
  int state = STATE_MOVE_FORWARD;

  wb_robot_init();
  
  init();
 
  wb_robot_step(TIME_STEP);

  image1 = (unsigned char *) malloc(sizeof(unsigned char)*width*height*8);

  printf("Transiting to state MOVE_FORWARD\n");
  
  while (wb_robot_step(TIME_STEP) != -1) {

    // grab the image, segment it into red pixels and return the number of
    // red pixels	  
    const unsigned char *image = wb_camera_get_image(camera);
    int nr_red = imageSegment(image,image1,width,height);

    WbImageRef iref = wb_display_image_new(display1, width, height, image1, WB_IMAGE_BGRA);
    wb_display_image_paste(display1, iref, 0, 0, false);  
  
    switch(state) {
      case STATE_MOVE_FORWARD:
        omegaL = 0.5*MAX_SPEED;
        omegaR = 0.5*MAX_SPEED;
        alpha = 0.0;
        omegaRobot = 0.0; 
        // If we see 10 or more red pixels then there is an object in front of ut
        if(nr_red > 10) {
          state = STATE_OBSTACLE_AVOID;
          printf("Transiting to state OBSTACLE_AVOID\n");
        }
        break;
      case STATE_OBSTACLE_AVOID:
        printf("\n");
        // Get the centre and height of the biggest blob
        int centre = getBiggestRedBlob(image1,width,height,&blobHeight);
        printf("centre = %d bHeight = %f\n",centre,blobHeight);
        // Calculate the angular acceleration of the robot
        if(centre < width/2 - 4 && centre >0) {
          alpha = -1.0*blobHeight/height;
        }
        else if(centre > width/2 + 4) {
          alpha = 1.0*blobHeight/height;
        } else {
          state = STATE_MOVE_FORWARD;
           printf("Transit to STATE_MOVE_FORWARD\n");       
        }
        break;      
    }

    gain = 0.01;
    
    // Update the robots angular velocity from its angular acceleration
    double desiredRadius = 0.25;
    omegaRobot += gain*alpha*TIME_STEP/1000;
    // And use this to calculate the wheel angular velocities
    omegaL = MAX_SPEED/10 + (desiredRadius - AXLE_LENGTH/2.0)*omegaRobot/WHEEL_RADIUS;
    omegaR = MAX_SPEED/10 + (desiredRadius + AXLE_LENGTH/2.0)*omegaRobot/WHEEL_RADIUS;    
    
    printf("Alpha = %f omegaRobot = %f\n",alpha,omegaRobot);
        
    wb_motor_set_velocity(left_motor, omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}
