// =================================================================
//
// CBP_ePuck_Camera_6.c 
//
// Linear Camera returns hue of what is in front on the ground
// CBP 19-11-20
//
// ==================================================================

#include <stdio.h>
#include <webots/camera.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/display.h>
#include <webots/distance_sensor.h>

#include <stdlib.h>

#include "CBP_Helper_4.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979
#define THRESHOLD_DIST 0.01

#define STATE_FRWD 1
#define STATE_TURN 2

static WbDeviceTag display1, display2;
WbDeviceTag toflight;
WbDeviceTag camera, left_motor, right_motor;
int width, height;
int widthD, heightD; //display
 
static unsigned char *image1;
static unsigned char *image2;

int getModeHue(const unsigned char *image, int width, int height) {
  
  int count = 0;
  double minsat = 1.5;
  double maxsat = 0;
  int hueArray[360];
  int maxHue = -1;
  double sat;
  
    for(int i=0; i< 360; i++)
      hueArray[i] = -1;
      
  for (int i = width/2-2; i < width/2+2 ; i++) {
    for (int j = 0; j < height; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      double green = wb_camera_image_get_green(image, width, i, j);
      double blue = wb_camera_image_get_blue(image, width, i, j);  
  
      //printf("r = %f, g = %f, b = %f\n",red,green,blue);
            
      // Find and remove white light pollution      
      double min = 300;
      if(red < min)
        min = red;
      if(green < min)
        min = green;
      if(blue < min)
        min = blue;

       double intens = red + green + blue + 0.000001;             
      // printf("intens = %f, min = %f,\n",intens,min);
       sat = (intens - min)/intens;
       if(sat < minsat)
         minsat = sat;
       if(sat > maxsat)
         maxsat = sat;
              
      red -= min;
      blue -= min;
      green -= min;                
               
      double sum = red + green + blue + 0.000001;
      red = red/sum;
      green = green/sum;
      blue = blue/sum;
         
      // printf("r = %f, g = %f, b = %f\n",red,green,blue);
      
      double numi=0.5*((red - green)+(red - blue));
      double denom=sqrt((red-green)*(red - green)+ (red-blue)*(green - blue));
      double hue = acos(numi/(denom + 0.0000001))*180/M_PI;
      
      if(blue > green) hue = 360-hue; 
      //printf("Hue = %f\n",hue);  
      hueArray[(int)hue]++;
      count++;
    }      
  }
    
  // Find the mode in the hue array
  maxHue = -1;
  int foundIndex = -1;
  for(int i=0; i< 360; i++){
    int hue;
    hue = hueArray[i];
    if(hue > maxHue) {
      maxHue = hue;
      foundIndex = i;
    }
  }
  if(sat < 0.7) foundIndex = -1;
  return foundIndex; 
}

int processBlob(const unsigned char *image,unsigned char *imageThresh,
                unsigned char *image2, int width, int height) {
  
  int count = 0;
  for (int i = width/2-2; i < width/2+2 ; i++) {
    for (int j = height/2-2; j < height/2+2; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      double green = wb_camera_image_get_green(image, width, i, j);
      double blue = wb_camera_image_get_blue(image, width, i, j);   
      // Find and remove white light pollution
      double min = 300;
      if(red < min)
        min = red;
      if(green < min)
        min = green;
      if(blue < min)
        min = blue;
      red -= min;
      blue -= min;
      green -= min; 
      // Now Saturation = 1  
      *(image2 + i*4 + 1 + j*width*4) = (unsigned char)red;
      *(image2 + i*4 + 2 + j*width*4) = (unsigned char)green;
      *(image2 + i*4 + 3 + j*width*4) = (unsigned char)blue;                
      count++;
      printf(" | %f  %f  %f",red,green,blue); 
    }
    printf("\n");
   }
   printf(" == \n");
    
  return count;
  
}

int getAvRedLocation(unsigned char *image, int width, int height){

   int indexSum = 0;
   int indexAv = 0;
   int count = 0;
   for (int i = 0; i < width ; i++) { 
     if(*(image + i*4 + 1 + (height/2)*width*4) == 250) {
       indexSum += i; 
       count++;
     }
   }
   printf("IndexSum = %d\n",indexSum);
   indexAv = indexSum/count;
   return indexAv;   
} 

void init() {

  camera = wb_robot_get_device("camera");
  wb_camera_enable(camera, TIME_STEP);
  width = wb_camera_get_width(camera);
  height = wb_camera_get_height(camera);
  printf("Camera width: %d height: %d\n",width,height);

  /* get a handler to the motors and set target position to infinity (speed control). */
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  // get the display device
  display1 = wb_robot_get_device("display");
  widthD = wb_display_get_width(display1);
  wb_display_set_font(display1,"Ariel",12,false);   

  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP);
 
}


int main() {

  double omegaL, omegaR;
  int count = 0;
  int state = STATE_FRWD;
  char str[10];
  int modeHueLast = -1;
  
  wb_robot_init();

  init();
 
  wb_robot_step(TIME_STEP);
  
  width = wb_camera_get_width(camera);
  height = wb_camera_get_height(camera);
  
  while (wb_robot_step(TIME_STEP) != -1) {
	  
    /* Get the new camera values */
    const unsigned char *image = wb_camera_get_image(camera);
 //   int nr_red = imageSegment(image,image1,width,height);
 //   int ret = processBlob(image,image1,image2,width,height);
    int modeHue = getModeHue(image,width,height);
    if(modeHue != -1)
      printf("Mode Hue = %d\n",modeHue);
      // Send color index to display when color changes.
    if(modeHue != modeHueLast){
      sprintf(str," ");

      wb_display_set_color(display1,0x0);
      wb_display_fill_rectangle(display1,0,0,64,16);       
      sprintf(str,"%3d",modeHue);
      wb_display_set_color(display1,0xFFFFFF);
      wb_display_draw_text(display1,str,1,1);    
      modeHueLast = modeHue;    
    } 
  
    double tof_value = wb_distance_sensor_get_value(toflight); 
    
    switch(state) {
      case STATE_FRWD:
        omegaL =  0.5*MAX_SPEED;
        omegaR =  0.5*MAX_SPEED;
        if(tof_value < 200) {
          printf("Transit to STATE_TURN\n");
          state = STATE_TURN;                    
        }
        break;

      case STATE_TURN:
        omegaL =  -0.5*MAX_SPEED;
        omegaR =  0.5*MAX_SPEED;
        bool done = delay(1.05);
        if(done) {
          done = false;
          state = STATE_FRWD;
          printf("Transit to STATE_FRWD\n");          
        }
        break;        
    }

    wb_motor_set_velocity(left_motor, omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
  }

  wb_robot_cleanup();

  return 0;
}

