// ==================================================================
//
// CBP_ePuck_LinearCamera_1.c
//
// 24-10-20
//
//
// ==================================================================

// Included libraries
#include <stdlib.h>
#include <stdio.h>
#include <webots/camera.h>
#include <webots/motor.h>  //obtain dif. wheels library
#include <webots/robot.h>  //obtain main library of webots
#include <webots/display.h>

// Global defines
#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979
#define THRESHOLD_DIST 0.01

WbDeviceTag cam;
static WbDeviceTag display1;
unsigned short width, height;
WbDeviceTag left_motor, right_motor;

int imageSegment(const unsigned char *image,unsigned char *imageThresh, int width, int height) {
  
  int count = 0;
  double minsat = 1.5;
  double maxsat = 0;
  for (int i = 0; i < width ; i++) {
    for (int j = 0; j < height; j++) {
      double red = wb_camera_image_get_red(image, width, i, j);
      double green = wb_camera_image_get_green(image, width, i, j);
      double blue = wb_camera_image_get_blue(image, width, i, j);  
      
      // Find and remove white light pollution
      
      double min = 255;
      if(red < min)
        min = red;
      if(green < min)
        min = green;
      if(blue < min)
        min = blue;
       double intens = red + green + blue + 0.000001;
       
       double sat = (intens - min)/intens;
       if(sat < minsat)
         minsat = sat;
       if(sat > maxsat)
         maxsat = sat;         
      red -= min;
      blue -= min;
      green -= min;
                     
      double sum = red + green + blue + 0.000001;
      red = red/sum;
      green = green/sum;
      blue = blue/sum;
      
      double numi=0.5*((red - green)+(red - blue));
      double denom=sqrt((red-green)*(red - green)+ (red-blue)*(green - blue));
      double hue = acos(numi/(denom + 0.0000001))*180/M_PI;
      if(blue > green) hue = 360-hue;   
      //printf("Sat = %f\n",sat);
      // 4 bytes per pixel in the image. Offset 1 is red
      *(imageThresh + i*4 + j*width*4) = 0;
      *(imageThresh + i*4+1 + j*width*4) = 0;
      *(imageThresh + i*4+2 + j*width*4) = 0;
      *(imageThresh + i*4+3 + j*width*4) = 0;      
      if(( hue > 50 && hue < 70 && sat > 0.7) ) {//|| (hue > 330 && hue < 360))
        *(imageThresh + i*4 + 1 + j*width*4) = 250;
        count++;
      }
    }
        
  
  }
  printf("Min sat = %f, max sat = %f\n",minsat,maxsat);
  return count;
  
}


int getAvLocation(unsigned char *image, int width, int height){

   //int row = height/2;
   int indexSum = 0;
   int indexAv = 0;
   int count = 0;
   for (int i = 0; i < width ; i++) { 
     if(*(image + i*4 + 1 + (height/2)*width*4) == 250) {
       indexSum += i; 
       count++;
     }
   }
 //  printf("IndexSum = %d\n",indexSum);
   if(count >0) indexAv = indexSum/count;
   return indexAv;   
} 

void init(){

  // get a handler to the motors and set target position to infinity (speed control)
  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  cam = wb_robot_get_device("camera");
  wb_camera_enable(cam, TIME_STEP);
  width = wb_camera_get_width(cam);
  height = wb_camera_get_height(cam);
  printf("Camera width = %d, height = %d\n",width,height);
  // get the display device
  display1 = wb_robot_get_device("display");

}

int main() {

  double omegaL, omegaR;
  wb_robot_init();
  init();

  unsigned char *lineImage = (unsigned char *)malloc(sizeof(unsigned char) * width * height * 8);

  while (wb_robot_step(TIME_STEP) != -1) {

    const unsigned char *image;

    // 1. Get the sensors values
    image = wb_camera_get_image(cam);

    int rn = imageSegment(image,lineImage, width, height);
    printf("rn = %d\n",rn);
    WbImageRef iref = wb_display_image_new(display1, width, height, lineImage, WB_IMAGE_BGRA);
    wb_display_image_paste(display1, iref, 0, 0, false);
    
    // 2. Generate control signal from image
    double loc = (double)getAvLocation(lineImage,width,height);
    
	// >>>>>>>>>>>>>>>>>>>>> Here you must calculate the error
    double gain = 4.0;
    //double error = 
    //printf("loc = %f, error = %f\n",loc,error);    
    
	// >>>>>>>>>>>>>>>>>>>>> Here you must use the error to calculate omegaL, omegaR
	
   // omegaL = 
   // omegaR = 

    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);
        
  }
  free(lineImage);
  wb_robot_cleanup();

  return 0;
}
