// ===================================================================
//
// FBO_ePuck_Maze_5.c
// CBP_ePuck_Maze_5
//
// Maze Follower
// CBP FBO 03-11-20
//
// ==================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <webots/distance_sensor.h>
#include <webots/position_sensor.h>
#include <stdio.h>
#include <webots/gps.h>
#include <webots/compass.h>

#include <windows.h>

#include "CBP_Plotter_2.h"
#include "CBP_Helper_4.h"

#define TIME_STEP 16
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

#define DIST_THRESHOLD 0.06

#define STATE_STEP 1
#define STATE_DECIDE 2
#define STATE_TURN_CLOCKWISE 3
#define STATE_TURN_ANTICLOCKWISE 4
#define STATE_TURN_ABOUTFACE 5
#define STATE_INTERMED 6
#define STATE_STEP_OLD 7

#define DEBUG false

WbDeviceTag left_motor, right_motor;
WbDeviceTag compass;
WbDeviceTag gps;
WbDeviceTag toflight;

// distance sensors
char ps_names[8][4] = {"ps0", "ps1", "ps2", "ps3", "ps4", "ps5", "ps6", "ps7"};
WbDeviceTag ps[8];
double ps_vals[8];
double leftDistSensor, rightDistSensor, forwardLeftSensor, forwardRightSensor;

WbDeviceTag leftWheelAngle,rightWheelAngle;

void init() {

  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
  
  compass = wb_robot_get_device("comp1");
  wb_compass_enable(compass, TIME_STEP);    
  
  // initialise infra-red distance  sensors
  for (int i = 0; i < 8; i++) {
    ps[i] = wb_robot_get_device(ps_names[i]);
    wb_distance_sensor_enable(ps[i], TIME_STEP);  
  }   
  
  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP); 
  
  // Initialize Wheel Encoders
  leftWheelAngle = wb_robot_get_device("left wheel sensor");
  wb_position_sensor_enable(leftWheelAngle,TIME_STEP);
  rightWheelAngle = wb_robot_get_device("right wheel sensor");
  wb_position_sensor_enable(rightWheelAngle,TIME_STEP);  

}

void driveMotors(double omL, double omR) {
  wb_motor_set_velocity(left_motor,  omL);
  wb_motor_set_velocity(right_motor, omR);
}

int getWallCode(){
  leftDistSensor  = wb_distance_sensor_get_value(ps[5]);
  rightDistSensor = wb_distance_sensor_get_value(ps[2]); 
  double leftDist = lookUpDistance(c,leftDistSensor);
  double rightDist = lookUpDistance(c,rightDistSensor);
  double centreDistSensor = wb_distance_sensor_get_value(toflight); 
  // Tale location of tof sensor on the ePuck
  double centreDist = 0.046 + lookUpDistanceTOF(cTOF,centreDistSensor);
  int code = 0;
  if(leftDist  != -1 && leftDist  < DIST_THRESHOLD) code += 4;
  if(rightDist != -1 && rightDist < DIST_THRESHOLD) code += 2;
 // if(centreDist  < 0.112/2.0*1.105) code += 1;
  if(centreDist  < 0.112/2.0*1.0) code += 1;   
  if(DEBUG)printf("Dists: L = %f, R == %f, C = %f\n",leftDist,rightDist,centreDist);
  return code;
}

int main(int argc, char **argv) {

  double omegaL, omegaR;
  int state;
  int lastState=0;
  int lastCode = 0;
  int code;
  double  dist;  
  double wheelThetaInit;
  double wheelTheta;
  double desiredDist;
  double distScaling;
  
  wb_robot_init();
  
  init();

  printf("Entering STATE_STEP\n");
  lastCode = 0;
  state = STATE_STEP;
  dist = 0;
  distScaling = 1.0;   
//  while (wb_robot_step(TIME_STEP) != -1) {
  while (1) {
                    
    switch(state) {         
 
       case STATE_STEP:
      
        //wb_robot_step(TIME_STEP); 
        lastCode = code;
        code = getWallCode();
        wb_robot_step(TIME_STEP);     
          
        double wheelThetaInit = wb_position_sensor_get_value(leftWheelAngle);
        dist = 0;
        // Below applies your odometry correction
        desiredDist = 0.112*distScaling;
        
        // WHile loop for move forward a step
                      
               
        state = STATE_DECIDE;  
        printf("Transit to STATE_DECIDE dist = %f\n",dist);
        printf("Last code = %d, code = %d\n",lastCode,code);        
        break;    
        
      case STATE_DECIDE:
        code = getWallCode();
        wb_robot_step(TIME_STEP);
         
        // selection of state to transit to for codes        



        printf("STATE_DECIDE: next state = %d\n",state);
        Sleep(500);
        break;
           
      case STATE_TURN_CLOCKWISE:
        printf("\n");
        wheelThetaInit = wb_position_sensor_get_value(leftWheelAngle);
        dist = 0;
        double desiredDist = (AXLE_LENGTH/2.0)*(PI/2);
        if(DEBUG)printf("Desired angle = %f\n",(desiredDist/(AXLE_LENGTH/2))*180/PI);
        
        // Correction from Odometry program
        desiredDist *= (90.7419/82.4567);        
        omegaL = 0.2*MAX_SPEED;
        omegaR = -omegaL;
        
        // while loop for turn clockwise
        
        state = STATE_STEP; 
        printf("Transit to STATE_STEP from CLOCKWISE\n"); 
        break;  
        
      case STATE_TURN_ANTICLOCKWISE:
        printf("\n");
        wheelThetaInit = wb_position_sensor_get_value(rightWheelAngle);
        dist = 0;
        desiredDist = (AXLE_LENGTH/2.0)*(PI/2);
        if(DEBUG)printf("Desired angle = %f\n",(desiredDist/(AXLE_LENGTH/2))*180/PI);
        
        // Change the below for your correction factor
        desiredDist *= (90.7419/82.4567);  
        omegaL = -0.2*MAX_SPEED;
        omegaR = -omegaL;

        // While loop for turn anticlockwise
        
               
        state = STATE_STEP; 
        printf("Transit to STATE_STEP from ANTICLOCKWISE\n");      
        break;  
        
      case STATE_TURN_ABOUTFACE:
        printf("\n");
        wheelThetaInit = wb_position_sensor_get_value(rightWheelAngle);
        dist = 0;
        desiredDist = (AXLE_LENGTH/2.0)*(PI);
        if(DEBUG)printf("Desired angle = %f\n",(desiredDist/(AXLE_LENGTH/2))*180/PI);
        
        // Change below with your correction factor
        desiredDist *= (90.7419/82.4567);          
        omegaL = -0.2*MAX_SPEED;
        omegaR = -omegaL;

        // while loop for about face    



		
        state = STATE_STEP; 
        printf("Transit to STATE_STEP from ABOUT FACE\n");        
        break;                           
    }     
  }

  wb_robot_cleanup();

  return 0;
}
