// =================================================
//
// CBP_ePuck_ObjDet2a.c
//
// 09-10-20                   RELEASE
// Based on CBP_ePuck_2e.c
//
// Rotate robot and scan for obstacles.Assume there 
// are 2 obstacles. Robot gets their thetas and
// finds the average, which is where it will move to
//
// ==================================================

#include <webots/distance_sensor.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/compass.h>
#include <webots/position_sensor.h>

#include <stdio.h>

#include "CBP_Polar_Plotter.h"
#include "CBP_Helper_4.h"

// time in [ms] of a simulation step
#define TIME_STEP 32
#define MAX_SPEED 6.28
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define DIST_THRESHOLD 0.5
#define TIME_TO_MOVE 5.0
// FSM
#define STATE_SCANNING 1
#define STATE_ANALYSING 2
#define STATE_ROTATETO_GAP 3
#define STATE_FORWARDTO_GAP 4
#define STATE_DONE 5

WbDeviceTag toflight;
WbDeviceTag compass;
WbDeviceTag right_motor;
WbDeviceTag left_motor;
WbDeviceTag leftWheelAngle;
  
// Initialize sensors and actuators
void init() {

  toflight = wb_robot_get_device("tof");
  wb_distance_sensor_enable(toflight,TIME_STEP);
  
  right_motor = wb_robot_get_device("right wheel motor");
  left_motor = wb_robot_get_device("left wheel motor");

  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
      
  compass = wb_robot_get_device("comp1");
  wb_compass_enable(compass, TIME_STEP);

  leftWheelAngle = wb_robot_get_device("left wheel sensor");
  wb_position_sensor_enable(leftWheelAngle,TIME_STEP);
  
}

int main(int argc, char **argv) {
  
  double theta;
  double omegaL = 0.0;
  double omegaR = 0.0;
  int thetaArray[360];
  double distArray[360];
  int thetaIndex=0;
  int state;
  int count = 0; 
  double avtheta=0;
  double avDist=0;
  double sumtheta = 0;
  double wheelTheta=0;
  double wheelThetaStart=0;
  int riseIndex, fallIndex;
  
  FILE* logFile;
  logFile = fopen("Log.m", "w");
  writeOctaveHead(logFile);
    
  wb_robot_init();

  init();
   
  for(int i=0;i<360;i++) {
    thetaArray[i] = 0;
  }
    
  state = STATE_SCANNING;
  
  while (wb_robot_step(TIME_STEP) != -1) {
  
    switch(state) {
      case STATE_SCANNING:
        theta = getAngleDegrees(compass);

        printf(">>> SCANNING: theta = %f  ",theta);        
        omegaL = -0.1*MAX_SPEED;
        omegaR = 0.1*MAX_SPEED;
        double sensorVal = wb_distance_sensor_get_value(toflight);
        double dist = lookUpDistanceTOF(cTOF,sensorVal);
        
        // Convert the float theta into int thetaIndex
        // then if < 0 add 360
        
        // If dist is less than DIST_THRESHOLD set thetaArray
        // at thetaIndex to 1.0 else to 0.0

        // Store the dist in the distArray at this index
 
                
        printf("thetaIndex %d, dist %f\n",(int)thetaIndex,dist); 
        writeData2ToOctave(logFile,(int)theta,dist);           
        if(theta < 1 && theta > -1)
          state = STATE_ANALYSING;
        break;
      
      case STATE_ANALYSING:
        printf(">>>>> STATE_ANALYSING\n");
        omegaL = 0.0;
        omegaR = 0.0;
        count = 0;
        sumtheta = 0; 
        fallIndex = 359;
        riseIndex = 0;

        for(int i=0;i<359;i++){
         // Uncomment and add code to find the location of the falling edge in thetaArray
         // if(  ){
         //   fallIndex = i;
         //   printf("Fall index = %d\n",fallIndex); 
         // }
         
         // Uncomment and add code to find the location of the risingedge in thetaArray
         // if(  ){
         //   riseIndex = i;
         //   printf("Rise index = %d\n",riseIndex);             
         // }
          
        }
        avtheta = (riseIndex + fallIndex)/2.0;
        printf(">>> avtheta = %f\n",avtheta);
        printf("Rise index = %d, fall index = %d\n",riseIndex,fallIndex);
        
        // calc dist to rise and fall points, and get average.
        // bump indices so we are not at the edge of the object
        double dist1 = distArray[riseIndex+2];
        double dist2 = distArray[fallIndex-2];
        avDist = (dist1 + dist2)/2.0;
        printf("dist1 = %f, dist2 = %f, avDist = %f\n",dist1,dist2,avDist);
        printf("theta1 = %d, theta2 = %d\n",fallIndex,riseIndex);
        
        // calc coordinates of both obstacle edges
        double x1 = dist1*cos(fallIndex*M_PI/180);
        double y1 = dist1*sin(fallIndex*M_PI/180);
        printf("(x1, y1) = (%f, %f) \n", x1,y1);
        
        double x2 = dist2*cos(riseIndex*M_PI/180);
        double y2 = dist2*sin(riseIndex*M_PI/180);
        printf("(x2, y2) = (%f, %f) \n", x2,y2);
        
        // calc coordinates of gap
        double xC = (x1+x2)/2;
        double yC = (y1+y2)/2;
        printf("(xC, yC) = (%f, %f) \n", xC,yC);               
        state = STATE_ROTATETO_GAP;
       
        break;   
                   
      case STATE_ROTATETO_GAP:      
        omegaL = -0.1*MAX_SPEED;
        omegaR = 0.1*MAX_SPEED;
          theta = getAngleDegrees(compass);             
        if(fabs(theta - avtheta) < 2.0) {
          printf("Final theta = %f\n",theta);
          state = STATE_FORWARDTO_GAP;
          // get wheel encoder value before starting to move
          wheelThetaStart = wb_position_sensor_get_value(leftWheelAngle);
          omegaL = 0.5*MAX_SPEED;
          omegaR = 0.5*MAX_SPEED;

          printf("avDist = %f omegas = %f\n",avDist,omegaL);
        }//
        break;
               
      case STATE_FORWARDTO_GAP:
        printf("\n"); // To fool compiler
        // Get encoder value and use to calculate perceived distance moved.
        wheelTheta = wb_position_sensor_get_value(leftWheelAngle) - wheelThetaStart;
        dist = wheelTheta*WHEEL_RADIUS;
        
        if(dist >= avDist) {
          omegaL = 0.0;
          omegaR = 0.0;           
          state = STATE_DONE;    
        }
        break; 
        
      case STATE_DONE:
        break;               
    }
    
   wb_motor_set_velocity(left_motor,  omegaL);
   wb_motor_set_velocity(right_motor, omegaR);
   
  }
 
  writeOctaveTail(logFile);
  fclose(logFile);
  // cleanup the Webots API
  wb_robot_cleanup();
  return 0;  // EXIT_SUCCESS
}
