// ===================================================================
// File:          CBP_ePuck_Odometry_3
// Date:          03-10-20 Release
//
// Rotates on an arc of given radius for given angle. Checks perceived
// location using wheel encoders with actual location using GPS
//
// ===================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <webots/distance_sensor.h>
#include <webots/position_sensor.h>
#include <stdio.h>
#include <webots/gps.h>
#include <webots/compass.h>

#include "CBP_Plotter_2.h"
#include "CBP_Helper_2.h"

#define TIME_STEP 32
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

#define STATE_CALC_OMEGAS 1
#define STATE_ROTATE 2
#define STATE_REPORT 3
#define STATE_DONE 4

WbDeviceTag left_motor, right_motor;
WbDeviceTag comp;
WbDeviceTag gps;

void init() {

  left_motor = wb_robot_get_device("left wheel motor");
  right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);

  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
  
  comp = wb_robot_get_device("comp1");
  wb_compass_enable(comp, TIME_STEP);    

}


int main(int argc, char **argv) {

  double omegaL, omegaR;
  int state;
  
  double bearingRobot;
  double angleRobot;

  double desiredRadius = 0.5;
  double desiredDegrees = 90;
  double desiredTime = 10.0;
  
  double robotOmega;
  double angleFromGPS;
  
  double wheelThetaL, wheelThetaR;
  double distL_perceived, distR_perceived;
  double distL_desired, distR_desired;
  double distC_desired, distC_perceived;
  double x_perceived, y_perceived;
  
  FILE* plotFile;
  plotFile = fopen("Robot.m", "w");
  writeOctaveHead(plotFile);  
  
  WbDeviceTag leftWheelAngle,rightWheelAngle;
  
  wb_robot_init();
  
  init();

  // Enable the wheel encoders here
  // leftWheelAngle = 
  // 
  // rightWheelAngle = 
  // wb_position_sensor_enable(rightWheelAngle,TIME_STEP);
    
    
  state = STATE_CALC_OMEGAS;
  
  while (wb_robot_step(TIME_STEP) != -1) {
  
    switch(state) {

      case STATE_CALC_OMEGAS:
        // Calculate angular velocities omegaL and omegaR
        // robotOmega = 
        // omegaR = 
        // omegaL = (  
        printf("omegaL = %f, omegaR = %f\n",omegaL,omegaR);
        
        // Calculate the desired distances the wheels must move and robot centre
         
        state = STATE_ROTATE;   
        break;
        
      case STATE_ROTATE:
        printf("\n");
        
        // Code to get the left wheel angle (wheelThetaL) and calculate the
        // perceived distance moved by the left wheel

        // Repeat for the right wheel, then the robot centre
 
        // You get the following lines for free. Calculates the (x,y) location
        // of the robot centre from the perceived centre distance
        x_perceived = desiredRadius*sin(distC_perceived/desiredRadius);
        y_perceived = desiredRadius*(1 - cos(distC_perceived/desiredRadius));       
 
        // get the GPS values, the exact robot (x,y) location
        const double *gpsVals = wb_gps_get_values(gps); 
                    
        // And write to an Octave file
        writeData4ToOctave(plotFile,x_perceived,y_perceived, gpsVals[2],gpsVals[0]);
             
        // Use distC_perceived to halt -------------------              
//        if(    ) {
//
//        }        
                      
        break; 
        
       case STATE_REPORT:  // Leave all of this for students
        
        printf("\n"); // To fool compiler             
        wheelThetaL = wb_position_sensor_get_value(leftWheelAngle);
        distL_perceived = wheelThetaL*WHEEL_RADIUS;
        wheelThetaR = wb_position_sensor_get_value(rightWheelAngle);
        distR_perceived = wheelThetaR*WHEEL_RADIUS;
        distC_perceived = (distL_perceived + distR_perceived)/2;
                                
        printf("Dist left (desired) = %f, (perceived) = %f\n",distL_desired,distL_perceived);
        printf("Dist right (desired) = %f, (perceived) = %f\n",distR_desired,distR_perceived);
        printf("Dist centre (desired) = %f, (perceived) = %f\n\n",distC_desired,distC_perceived);
               
        // Calculate perceived coords
        x_perceived = desiredRadius*sin(distC_perceived/desiredRadius);
        y_perceived = desiredRadius*(1 - cos(distC_perceived/desiredRadius)); 

        printf("Robot perceived  coords: = (%f,%f)\n",x_perceived,y_perceived);
        printf("Robot gps actual coords: x = %f , y = %f\n\n",gpsVals[2],gpsVals[0]);   
        
        // Get robot angle - perhaps not give this to studs.
        bearingRobot = getBearing(comp);
        angleRobot = 180.0/2.0 - bearingRobot;
        if(angleRobot < -180)
          angleRobot += 360;
        angleFromGPS = (180/PI)*atan2(gpsVals[0],gpsVals[2]);             
        printf("angleRobot = %f 2*angleFromGPS = %f\n",angleRobot,2*angleFromGPS); 
                    
        state = STATE_DONE;
        break;
      case STATE_DONE:
        omegaL = 0.0;
        omegaR = 0.0;  
        break;      
    
    }

    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);

  };

  writeOctaveTail(plotFile);
  fclose(plotFile);
    
  wb_robot_cleanup();

  return 0;
}
