// ===================================================================
// File:          CBP_ePuck_Odometry_5a
// Date:
// 30-10-20
//
// Move forward a distance, odometry calibration
//
// ==================================================================

#include <webots/robot.h>
#include <webots/motor.h> 
#include <webots/distance_sensor.h>
#include <webots/position_sensor.h>
#include <stdio.h>
#include <webots/gps.h>
#include "CBP_Plotter.h"
#include "CBP_Helper_2.h"

#define TIME_STEP 16
#define WHEEL_RADIUS 0.0205
#define AXLE_LENGTH 0.053
#define MAX_SPEED 6.28
#define PI 3.14159265358979

#define STATE_MOVE 1
#define STATE_CALCULATE 2
#define STATE_DONE 3


int main(int argc, char **argv) {

  double omegaL, omegaR;
  double correctionFactor;
  
  FILE* plotFile;
  plotFile = fopen("Robot.m", "w");
  writeOctaveHead(plotFile);  
  
  WbDeviceTag leftWheelAngle, rightWheelAngle;
  
  wb_robot_init();

  WbDeviceTag left_motor = wb_robot_get_device("left wheel motor");
  WbDeviceTag right_motor = wb_robot_get_device("right wheel motor");
  wb_motor_set_position(left_motor, INFINITY);
  wb_motor_set_position(right_motor, INFINITY);
  wb_motor_set_velocity(left_motor, 0.0);
  wb_motor_set_velocity(right_motor, 0.0);
  WbDeviceTag gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP);   
  
  leftWheelAngle = wb_robot_get_device("left wheel sensor");
  wb_position_sensor_enable(leftWheelAngle,TIME_STEP);
  rightWheelAngle = wb_robot_get_device("right wheel sensor");
  wb_position_sensor_enable(rightWheelAngle,TIME_STEP);
    
  while (1) {
    
    omegaL = 0.2*MAX_SPEED;
    omegaR = omegaL;
    
    //correctionFactor = 90.185/82.074;
    correctionFactor = 90.7419/82.4567;    
    correctionFactor = 1.0;
    
    double desiredDist = 0.112;// For the maze
    desiredDist *= correctionFactor;
    
    double wheelTheta = wb_position_sensor_get_value(rightWheelAngle);
    double dist = wheelTheta*WHEEL_RADIUS;
    wb_robot_step(TIME_STEP);
    
    while(dist < desiredDist) {
      wb_robot_step(TIME_STEP);
      wheelTheta = wb_position_sensor_get_value(rightWheelAngle);
      dist = wheelTheta*WHEEL_RADIUS;
      const double *gpsVals = wb_gps_get_values(gps);          
      printf("Desired dist = %f, dist = %f GPS:z = %f\n",desiredDist,dist,gpsVals[2]); 
      wb_motor_set_velocity(left_motor,  omegaL);
      wb_motor_set_velocity(right_motor, omegaR);	  
    }
    
    omegaL = 0;
    omegaR = 0;

    wb_motor_set_velocity(left_motor,  omegaL);
    wb_motor_set_velocity(right_motor, omegaR);

  };

  writeOctaveTail(plotFile);
  fclose(plotFile);
    
  wb_robot_cleanup();

  return 0;
}
